/* Dark Well - old-good game, mod of tetris.
 * Copyright (C) 2000 Boris Kozlov.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/time.h>
#include <curses.h>
#include "draw.h"
#include "dwtitn.h"
#include "bye.h"
#include "score.h"

struct pos {
         int x,y,l;
       };

typedef unsigned char vec[4];
typedef unsigned char matr[23][29];
typedef struct pos fordel[11];

vec clr;
matr glass;
unsigned char full=0,newg=0,lev=1,mpt,lr,ld,ldr,ldl,curturn;
unsigned char blscore,nextshown=1,snd=1;
int lastkey,score,total;

#define bl 219

void makebeep(int freq, int leng)
{
  if (snd) {
    fprintf(stdout,"\33[10;%d]\33[11;%d]\a\33[10]\33[11]",freq,leng);
    fflush(stdout);
  }
}

void init()
{
  srand(time(0));
  printf("\e(U"); /* set IBM_PC graphics set */      
  initscr();
  keypad(stdscr,1);
  meta(stdscr,1);
  noecho();   
  nonl();   /* don't wait for carriage return */
  cbreak();  
  clear();
  define_colours();
} //init

int titlescr()
{
  unsigned key=0,k;
  
  printf("%s",screendata); 
  endwin();

  curs_set(0);  
  cbreak();    
  while(key!=13 && key!=113 && key!=81 && key!=72 && key!=104) key=getchar();

  
  switch(key) {
    case 13 : k=1; break; //play
    case 81 :    
    case 113 : k=0; break; //quit   
    case 72 :   
    case 104 : k=2; break; //help
  }  
  return(k);
}

byescr()
{
  initscr();
  noecho();   /* shuttoff tty echoing */
  nonl();   /* don't wait for carriage return */
  cbreak();  
  curs_set(1);  	    
  printf("%s",byedata); 
  endwin();
}

void exiting(unsigned mode)
{
  clear();  
  printw("\e(B"); // set unix character set 
  if (!mode) byescr();   //0-normal quit
  exit(0);  
} //exiting


unsigned curdelay(int speed)
{
  return(65*(10-speed));
} //curdelay


unsigned long timer()
{
   struct timeval t;
   gettimeofday(&t,0);
   return ((unsigned long)t.tv_sec)*1000+t.tv_usec/1000;
} //timer

int waitkey(int delay)
{
   if (delay<0) delay=0;
   timeout(delay);
   lastkey=getch();
   timeout(-1);
#ifndef NCURSES_VERSION /* workaround for not working timeout(-1) */
   nocbreak(); cbreak();
#endif
   return lastkey;
} //waitkey

/*-----------------------------------------------------------------------*/

void setcolor(char col)
{
  switch(col)
  {
     case 2 : usecolor(COLOR_GREEN,A_NORMAL); break;
     case 9 : usecolor(COLOR_BLUE,A_BOLD); break;
     case 10 : usecolor(COLOR_GREEN,A_BOLD); break;
     case 11 : usecolor(COLOR_CYAN,A_BOLD); break;
     case 12 : usecolor(COLOR_RED,A_BOLD); break;
     case 13 : usecolor(COLOR_MAGENTA,A_BOLD); break;
     case 14 : usecolor(COLOR_YELLOW,A_BOLD); break;
     case 15 : usecolor(COLOR_WHITE,A_BOLD); break;
     case 0 : usecolor(COLOR_BLACK,A_NORMAL); break;
  }
} //setcolor

unsigned char askquit(unsigned char t)
{
  char *st1="Do you wanna quit to title screen ? (y/n)";
  char *st2="Do you wanna quit to shell? (y/n)";
  char *st;
  int key=0;
  unsigned char i;

  if (t==1) st=st1; else st=st2;
  move(24,10); clrtoeol();

  setcolor(15); centre(st,24); refresh();
  while(key!=121 && key!=110 && key!=89 && key!=78) key=getch();

  if (key==121 || key==89) return(1);//exit
  else { move(24,10); clrtoeol(); return(0); } //stay
} // askquit

void redraw(unsigned cl,vec col,int x,int y)
{
  char i;
  for (i=3;i>-1;i--)
    if (!((i==0) && (y1+y+i-2)==0)) {
      if (cl!=0) setcolor(col[i]); else setcolor(0);

      if (i==0) wrc2(' ',y1+y+i-2,x1+x-1);
      else if (cl!=0) wrc2(bl,y1+y+i-2,x1+x-1);
      else if (cl==0) wrc2(' ',y1+y+i-2,x1+x-1);
      refresh();
    }
} //redraw

void scrollu(vec col,int x,int y)
{
  char t;
  if (!(col[1]==col[2]==col[3])) {
    t=col[1]; col[1]=col[2]; col[2]=col[3]; col[3]=t;
    redraw(1,col,x,y);
  }
} //scrollu

void scrolld(vec col, int x,int y)
{
  char t;
  if (!(col[1]==col[2]==col[3])) {  
    t=col[3]; col[3]=col[2]; col[2]=col[1]; col[1]=t;
    redraw(1,col,x,y);
  }    
} //scrolld

void shad(char cl, int x)
{
  if (cl==0) setcolor(2); else setcolor(12);
  wrc2('',y1+height,x1+x-1);
  refresh();
} //shad

void makepause()
{
  char *st="Paused. Press 'p' to continue.";
  char key=0;

  setcolor(15); centre(st,24); refresh();
  while (key!=112 && key!=80) key=getch();
  move(24,10); clrtoeol(); refresh();
} //makepause

void fall(void)
{
  unsigned char x=width/2,y=1,i,j,land=0,drop=0;
  unsigned long end_time,start_time;
  vec col;

  for (i=1;i<4;i++) col[i]=glass[i][width/2]/10;
  for (i=1;i<4;i++) glass[i][width/2]=0;
  col[0]=0;
  shad(1,x);
  blscore=20;

  start_time=timer();
  for(;;) {
    end_time=start_time+curdelay(lev);
    for(;;) {
      if(timer()-start_time>=end_time-start_time) { blscore--; break; }
      switch(waitkey(end_time-timer())) {
      /* LEFT */
        case('j'):
        case('J'):
        case(KEY_A1):
        case(KEY_LEFT):
                        if (x > 1) {
                          if ((glass[y][x-2]==0) && (glass[y+1][x-2]==0) && (glass[y+2][x-2]==0)) {
                            redraw(0,col,x,y); shad(0,x);
                            x-=2;
                            redraw(1,col,x,y); shad(1,x);
                           }
                         } break;
        /* RIGHT */
        case('l'):
        case('L'):
        case(KEY_A3):
        case(KEY_RIGHT):
                         if (x < width-2) {
                           if ((glass[y][x+2]==0) && (glass[y+1][x+2]==0) && (glass[y+2][x+2]==0)) {
                             redraw(0,col,x,y); shad(0,x);
                             x+=2;
                             redraw(1,col,x,y); shad(1,x);
                           }
                         } break;
        /* UP */
        case('i'):
        case('I'):
        case(KEY_UP): scrollu(col,x,y); break;
        /* DOWN */
        case('k'):
        case('K'):
        case(KEY_DOWN): scrolld(col,x,y); break;
        /* DROP */
        case(13): land=1; end_time=timer(); break;
        case(' '): drop=1; break;
        /* PAUSE */
        case('p'):
        case('P'): flushinp(); makepause(); end_time=timer(); break;
        /* QUIT */
        case('Q'):
        case('q'):
        case(27): if (askquit(2)==1) exiting(0); else end_time=timer(); break;
	/* SOUND */
	case('S'):
	case('s'): setcolor(15); 
	           if (snd) { snd=0; wr("sound off",24,0); clrtoeol(); }
	           else  { snd=1; wr("sound on ",24,0); clrtoeol(); } break;
	/* SHOW NEXT*/	   
	case('N'):
	case('n'): setcolor(15); 
	           if (nextshown) {
		     nextshown=0; wr("next colors off",24,0); 
		     for (i=1;i<4;i++) { setcolor(15); wrc2('x',18+i,67); }      
		   }     
	           else { 
		     nextshown=1; wr("next colors on ",24,0);
                     for (i=1;i<4;i++) { setcolor(clr[i]/10); wrc2(bl,18+i,67); }		     
		   } break;
        /* INC LEV */		   
	case(43) : if (lev<9) lev++; setcolor(10); wrnum(lev,4,75); break;
	/* NEW GAME */
	case('G'):
	case('g'): newg=1; return; break;
        default: break;
      }
      if (land || drop) break;
    }
    if (drop) break;
    land=0;
    if (y+3 > height || glass[y+3][x]!=0) land=1;
    else { y++; redraw(1,col,x,y); }
    start_time=end_time;
    if (land) break;
  }

  if (drop) {
    while ((glass[y+3][x]== 0) && (y+3 <= height)) {
      redraw(0,col,x,y);
      y++;
      redraw(1,col,x,y);
      delay_output(1000);
      refresh();
    }
  }
  for (i=0;i<3;i++) glass[y+i][x]=col[i+1];
  if (x!=width/2) shad(0,x);
  move(24,0); clrtoeol();
} //fall

//---------------------------------------------------------------
void fallglass()
{
  unsigned char i,j,l;

  for
 (i=2;i<height+1;i++)
    for (j=1;j<width+1;j+=2)
      if ((glass[i][j]==0) && (glass[i-1][j]!=0)) {
        l=i-1;
        while (glass[l][j]!=0) {
          glass[l+1][j]=glass[l][j];
          setcolor(glass[l][j]); wrc2(bl,y1+l,x1+j-1);

          setcolor(0); wrc2(' ',y1+l-1,x1+j-1);
          glass[l][j]=0;
          if (l > 1) l--;
        }
        delay_output(4000);
      }
} //fallglass

void deld(char i,char j,char k)
{
  unsigned char l;

  for (l=0;l<k;l++) {
    glass[i+l][j]=0;
    setcolor(0); wrc2(' ',y1+i+l-1,x1+j-1);
  }
} //deld

void delr(char i,char j,char k)
{
  unsigned char l;

  for (l=0;l<k;l++) glass[i][j+l]=0;
  move(y1+i-1,x1+j-1);
  for (l=1;l<k/2+1;l++) {
    setcolor(0); printw("%c",' ');
 printw("%c",' ');
  }
} //delr

void deldl(char i,char j,char k)
{
  unsigned char l;

  for (l=0;l<k/2;l++) glass[i+l][j-l*2]=0;
  for (l=0;l<k/2;l++) {
    setcolor(0); wrc2(' ',y1+i+l-1,x1+j-1-l*2);
  }

} //deldl

void deldr(char i,char j,char k)
{
  unsigned char l;

  for (l=0;l<k/2;l++) glass[i+l][j+l*2]=0;
  for (l=0;l<k/2;l++) {
    setcolor(0); wrc2(' ',y1+i+l-1,x1+j-1+l*2);
  }
} //deldr

char check()
{
  unsigned i,j,k1,k2,k3,k4;
  unsigned was1,was2,was3,was4,was=0;
  unsigned d1=1,d2=1,d3=1,d4=1;
  fordel dwn,rgt,dlft,drgt;

  for (i=1;i<11;i++) dwn[i].l=rgt[i].l=dlft[i].l=drgt[i].l=0;

  for (i=1;i<height+1;i++)
    for (j=1;j<width+1;j+=2) {

       if (glass[i][j]!=0) {
         k1=0; k2=0; k3=0; k4=0;
         was1=0; was2=0; was3=0; was4=0;
         while ((glass[i+k1][j]==glass[i][j]) && (i+k1<=height)) k1++; // down
         while ((glass[i][j+k2]==glass[i][j]) && (j+k2 < width)) k2+=2; // right
         while ((glass[i+k3 / 2][j-k3]==glass[i][j]) && (i+k3 / 2 <= height) &&
         (j-k3 > 0)) k3+=2; // down-left
         while ((glass[i+k4 / 2][j+k4]==glass[i][j]) && (i+k4 / 2 <= height) &&
         (j+k4 <= width)) k4+=2; // down-right

         if ((k1 >=3) && !(glass[i][j]==glass[i-1][j])) { //down
           dwn[d1].x=j; dwn[d1].y=i; dwn[d1].l=k1;
           d1++;
           was1=1;
           score+=k1/2+blscore/3+abs(nextshown-1)*2; setcolor(10); wrnum(score,3,75);
           ld++; setcolor(2); wrnum(ld,9,75);
          }

         if ((k2 >=6) && !(glass[i][j]==glass[i][j-2])) { //right
           rgt[d2].x=j; rgt[d2].y=i; rgt[d2].l=k2;
	   d2++;
           was2=1;
           score+=1+k2/2+blscore/3+abs(nextshown-1)*2; setcolor(10); wrnum(score,3,75);
           lr++; setcolor(2); wrnum(lr,10,75);
         }

         if ((k3 >=6) && !(glass[i][j]==glass[i-1][j+2])) { //down-left
           dlft[d3].x=j; dlft[d3].y=i; dlft[d3].l=k3;
           d3++;
           was3=1;
           score+=3+k3/2+blscore/2+abs(nextshown-1)*4; setcolor(10); wrnum(score,3,75);
           ldl++; setcolor(2); wrnum(ldl,11,75);
         }

         if ((k4 >=6) && !(glass[i][j]==glass[i-1][j-2])){ //down-right
           drgt[d4].x=j; drgt[d4].y=i; drgt[d4].l=k4;
	   d4++;
           was4=1;
           score+=3+k4/2+blscore/2+abs(nextshown-1)*4; setcolor(10); wrnum(score,3,75);
           ldr++; setcolor(2); wrnum(ldr,12,75);
         }

         total=ldl+ldr+ld+lr;
	 setcolor(10); wrnum(total,13,75);

         if ((was1==1) || (was2==1) || (was3==1) || (was4==1)) {
           was=1;
           if ((was1 && was2) || (was1 && was3) || (was1 && was4) ||
           (was2 && was3) || (was2 && was4) || (was3 && was4)) {
             score+=20; setcolor(10); wrnum(score,3,75);
           }  
         }
       }
    }

  refresh();
  for (i=1;i<11;i++) {
    if (dwn[i].l!=0) {
      deld(dwn[i].y,dwn[i].x,dwn[i].l);
      refresh(); makebeep(6000,20);
      curturn++;
    }
    if (rgt[i].l!=0) {
      delr(rgt[i].y,rgt[i].x,rgt[i].l);
      refresh(); makebeep(6000,20);
      curturn++;      
    }
    if (dlft[i].l!=0) {
      deldl(dlft[i].y,dlft[i].x,dlft[i].l);
      refresh(); makebeep(6000,20);
      curturn++;      
    }
    if (drgt[i].l!=0) {
      deldr(drgt[i].y,drgt[i].x,drgt[i].l);
      refresh(); makebeep(6000,20);
      curturn++;      
    }
  }
  delay_output(6000);

  fallglass();
  refresh();

  return(was);
} //check
//---------------------------------------------------------------

void newbl()
{
  unsigned char i;

  full=0;
  if (!((glass[1][width/2]==0) && (glass[2][width/2]==0) &&
  (glass[3][width/2]==0))) full=1;

  if (full==0) for (i=1;i<4;i++) {
    glass[i][width/2]=clr[i];
    setcolor(clr[i]/10);
    wrc2(bl,y1+i-2+1, x1-1+width/2);
  }
} // newbl

void ncol(vec clr) //fill up clr and draw next colors
{
  unsigned char i;
  for (i=1;i<4;i++) clr[i]=(9+random()%6)*10;
  clr[0]=0;
  if (nextshown) 
    for (i=1;i<4;i++) { setcolor(clr[i]/10); wrc2(bl,18+i,67); }
  else    
    for (i=1;i<4;i++) { setcolor(15); wrc2('x',18+i,67); }  
  refresh();
} //ncol

void showinit()
{
  setcolor(2); wrnum(mpt,5,75); 
  wrnum(ld,9,75);
  wrnum(lr,10,75);
  wrnum(ldl,11,75);
  wrnum(ldr,12,75);
  setcolor(10);
  wrnum(total,13,75);  
  setcolor(10); wrnum(lev,4,75); wrnum(score,3,75);            
} //showinit

void clearwell()
{
  unsigned char i,j;
  for (i=y1;i<y1+height;i++) for (j=x1;j<x1+width;j++) wr(" ",i,j);
  refresh();    
} //clearwell

void showhelp()
{
  clear();

  setcolor(9);
  mvprintw(1,35,"Game Help");	  

  setcolor(2);  
  mvprintw(3,3,"Keys for moving blocks :");
  setcolor(10);    
  mvprintw(4,5,"left arrow,J,j");  
  mvprintw(5,5,"right arrow,L,l");    
  mvprintw(6,5,"up arrows,I,i");      
  mvprintw(7,5,"down arrow,K,k");        
  mvprintw(8,5,"Enter");          
  mvprintw(9,5,"Space");            

  setcolor(15);  
  mvprintw(4,23," - move block left");  
  mvprintw(5,23," - move block right");
  mvprintw(6,23," - scroll colors up");      
  mvprintw(7,23," - scroll colors down");        
  mvprintw(8,23," - slow drop: keep it pressed and see what it is");          
  mvprintw(9,23," - drop block");            

  setcolor(2);  
  mvprintw(11,3,"Pause/quit :");

  setcolor(10);  
  mvprintw(12,5,"P,p");
  mvprintw(13,5,"Q,q"); 
  mvprintw(14,5,"C-c"); 
  
  setcolor(15);    
  mvprintw(12,23," - pause game until this key pressed again");              
  mvprintw(13,23," - quit game with confirmation");                
  mvprintw(14,23," - quit game without it, as usual");                  

  setcolor(2);  
  mvprintw(16,3,"Options :");
  
  setcolor(10);    
  mvprintw(17,5,"N,n");                
  mvprintw(18,5,"S,s");                  
  mvprintw(19,5,"G,g");                    
  mvprintw(20,5,"+");                      
               
  setcolor(15);      
  mvprintw(17,23," - don't show/show next colors (way to get more scores)");            
  mvprintw(18,23," - turn off/on sound when deleting blocks");                    
  mvprintw(19,23," - start new game without confirmation");                      
  mvprintw(20,23," - increase game speed (it changes from 1 to 9)");                        
  
  setcolor(9);	  
  mvprintw(23,33,"Press any key.");	  
  setcolor(9);	  
  box(stdscr,0,0);	  
  getch();	  
} //showhelp

compareturn()
{
  if (curturn>mpt) mpt=curturn;
  setcolor(2); wrnum(mpt,5,75);
  
  if ((total/50 > lev) && !(lev==9)) { lev++; setcolor(10); wrnum(lev,4,75); }
} //compareturn

quickhelp()
{
  setcolor(2);
  wr("left,j  - move left",16,0);
  wr("right,l - move right",17,0);  
  wr("up,i    - scroll colors up",18,0);      
  wr("down,k  - scroll down",19,0);        
  wr("space   - drop block",20,0);    
  wr("q       - quit",21,0);
  wr("p       - pause game",22,0);
      
} //quickhelp

int main()
{
  unsigned i,j,k;

  init();
  openscore();
  while (k=titlescr()) {
    if (k==2) showhelp(); else {
      initscr();   
      drawwell();    
      quickhelp();
      printscore();    
      do {
        clearwell();
        for(i=0;i<width+1;i++) for(j=0;j<height+1;j++) glass[j][i]=0;

        ncol(clr);
        newg=score=0; lev=1; mpt=lr=ld=ldr=ldl=total=0;      
        showinit();
      
        do {
          full=0;
	  newbl();
          if (!full) { 
	    ncol(clr);
	    fall();
	    if (!newg) { curturn=0; while (check()); compareturn(); }
	  }  
        } while (!full && !newg); 
        if (score && !newg) savescore(score);
      } while (newg || !askquit(1));
    }
  }
  exiting(0);
  
}
